; NES controller reading code
; Copyright 2016 Johnathan Roatch
;

.include "nes.inc"

.importzp temp
.exportzp pad_1_keys, pad_1_axlr, pad_2_keys, pad_2_axlr

.export pads_read, pads_read_16bit, pads_rejectOpposingDirections

.segment "ZEROPAGE": zeropage
  pad_1_keys: .res 1
  pad_1_axlr: .res 1
  pad_2_keys: .res 1
  pad_2_axlr: .res 1

.segment "CODE"

;;
; Read both NES controller ports in big endian bit order.
; Also keeps in mind the even odd cycles after a DMA aliment.
; To avoid DMC glitches, call this immediately after the `sta OAM_DMA`.
; http://forums.nesdev.com/viewtopic.php?p=172188#p172188
;
; Preserves X, Y.
.proc pads_read
    ; 8 Bits from the controllers are shifted into pad_1 and pad_2.
  lda #$01
    ; In addition, $01 in pad_2_keys serves as the loop counter:
    ; once the bit gets shifted left eight times, it'll end up in carry
    ; terminating the read loop.
  sta pad_2_keys
  sta JOY_STROBE
  lsr
  sta JOY_STROBE
    ; Write 1 then 0 to JOY_PORT_1 to send a latch signal, telling the
    ; controllers to copy button states into a shift register.
    ;
    ; Also the strobe part must take an odd total number of cycles.
  read_loop:
      ; On NES and AV Famicom, button presses always show up in D0.
      ; On the original Famicom, presses on the hardwired controllers
      ; show up in D0 and presses on plug-in controllers show up in D1.
      ; D2-D7 consist of data from the Zapper, Power Pad, Vs. System
      ; DIP switches, and bus capacitance; so we ignore them.
    lda JOY_PORT_1    ; read player 1's controller
    and #%00000011    ; ignore D2-D7
    cmp #1            ; CLC if A=0, SEC if A>=1
    rol pad_1_keys    ; shift the the bit into memory
    bit $24
      ; 3 cycles with harmless flag setting side effects.
      ; So that the cycle parity is even on loading JOY_PORT_2
    lda JOY_PORT_2    ; read player 2's controller the same way
    and #%00000011
    cmp #1
    rol pad_2_keys
      ; The loop part must take an even number of cycles.
  bcc read_loop
    ; once $01 has been shifted 8 times, we're done
rts
.endproc

;;
; Read 16 both NES controller ports.
; Works the same as the 8 bit version above.
;
; Preserves X, Y.
.proc pads_read_16bit
  lda #$01
  sta pad_2_axlr
  sta JOY_STROBE
  eor pad_2_axlr
    ; Because we used a 3 cycle opcode to write a zero to
    ; the first byte of pad_2, we need to even the odds by
    ; using a odd cycle opcode that functions the same as `lda #0`
    ; XORing A with itself will clear all the bits.
  sta pad_2_keys
  sta JOY_STROBE
  read_loop:
    lda JOY_PORT_1
    and #%00000011
    cmp #1
    rol pad_1_axlr
    rol pad_1_keys
    lda JOY_PORT_2
    and #%00000011
    cmp #1
    rol pad_2_axlr
    rol pad_2_keys
    bit $24
      ; the extra 5 cycle rol already evens cycle parity.
      ; Here we even out the odd cycles in the loop branch.
  bcc read_loop
rts
.endproc

.proc pads_rejectOpposingDirections
  lda pad_1_keys
  jsr fix_byte
  sta pad_1_keys
  lda pad_2_keys
  jsr fix_byte
  sta pad_2_keys
rts
fix_byte:
    ; Construct a XOR mask for when up+down or left+right are pressed.
  sta temp+0
  asl
  and temp+0
    ; First apply bitwise AND to up+down and left+right.
  and #%00001010
    ; mask out other buttons.
  sta temp+1
  lsr
  ora temp+1
  ; Then dupplicate the resulting bits to each direction.
  cmp #BUTTON_UP|BUTTON_DOWN|BUTTON_LEFT|BUTTON_RIGHT
    ; If all directions are pressed simultaneously, then clear all keys.
    ; this is to prevent loose connectors from accidentally
    ; soft reseting (A+B+Start+Select), and so that "The Game Handler"
    ; won't be held upside-down.
  bne @dont_clear_all_keys
    lda temp+0
      ; loading the same value as the value to XOR will
      ; result in $00 being stored to temp+0
  @dont_clear_all_keys:
    ; Constructed mask now contains the buttons to toggle off.
  eor temp+0
rts
.endproc





